@@ -1,16 +1,18 @@
 import functools
+import gc
 import math
 import unittest
 from ctypes import Structure, byref, c_char, c_double, c_float, c_int, c_void_p, cast, create_string_buffer
 from decimal import Decimal
 from enum import Enum
+import weakref
 
 from rubicon.objc import (
     SEL, NSEdgeInsets, NSEdgeInsetsMake, NSMakeRect, NSObject, NSObjectProtocol, NSRange, NSRect, NSSize, NSUInteger,
     ObjCClass, ObjCInstance, ObjCMetaClass, ObjCProtocol, at, objc_classmethod, objc_const, objc_ivar, objc_method,
-    objc_property, send_message, send_super, types,
+    objc_property, py_from_ns, send_message, send_super, types,
 )
-from rubicon.objc.runtime import get_ivar, libobjc, objc_id, set_ivar
+from rubicon.objc.runtime import get_ivar, libobjc, objc_id, set_ivar, autoreleasepool
 
 from . import OSX_VERSION, rubiconharness
 
@@ -166,7 +168,6 @@ class RubiconTest(unittest.TestCase):
         random_obj = NSObject.alloc().init()
         with self.assertRaises(ValueError):
             ObjCClass(random_obj.ptr)
-        random_obj.release()
 
     def test_objcmetaclass_requires_metaclass(self):
         """ObjCMetaClass only accepts metaclass pointers."""
@@ -174,7 +175,6 @@ class RubiconTest(unittest.TestCase):
         random_obj = NSObject.alloc().init()
         with self.assertRaises(ValueError):
             ObjCMetaClass(random_obj.ptr)
-        random_obj.release()
 
         with self.assertRaises(ValueError):
             ObjCMetaClass(NSObject.ptr)
@@ -185,7 +185,6 @@ class RubiconTest(unittest.TestCase):
         random_obj = NSObject.alloc().init()
         with self.assertRaises(ValueError):
             ObjCProtocol(random_obj.ptr)
-        random_obj.release()
 
     def test_objcclass_superclass(self):
         """An ObjCClass's superclass can be looked up."""
@@ -616,12 +615,16 @@ class RubiconTest(unittest.TestCase):
         self.assertEqual(example.smiley(), "%-)")
 
     def test_number_return(self):
-        "If a method or field returns a NSNumber, it is converted back to native types"
+        "If a method or field returns a NSNumber, it is not automatically converted to a Python number."
         Example = ObjCClass('Example')
         example = Example.alloc().init()
 
-        self.assertEqual(example.theAnswer(), 42)
-        self.assertAlmostEqual(example.twopi(), 2.0 * math.pi, 5)
+        answer = example.theAnswer()
+        self.assertIsInstance(answer, ObjCInstance)
+        self.assertEqual(py_from_ns(answer), 42)
+        tau = example.twopi()
+        self.assertIsInstance(tau, ObjCInstance)
+        self.assertAlmostEqual(py_from_ns(tau), 2.0 * math.pi, 5)
 
     def test_float_method(self):
         "A method with a float argument can be handled."
@@ -662,8 +665,8 @@ class RubiconTest(unittest.TestCase):
         example = Example.alloc().init()
 
         result = example.areaOfTriangleWithWidth_andHeight_(Decimal('3.0'), Decimal('4.0'))
-        self.assertEqual(result, Decimal('6.0'))
-        self.assertIsInstance(result, Decimal, 'Result should be a Decimal')
+        self.assertIsInstance(result, ObjCClass('NSDecimalNumber'))
+        self.assertEqual(py_from_ns(result), Decimal('6.0'))
 
     def test_auto_struct_creation(self):
         "Structs from method signatures are created automatically."
@@ -974,15 +977,83 @@ class RubiconTest(unittest.TestCase):
         box.data = None
         self.assertIsNone(box.data)
 
+    def test_class_python_properties(self):
+
+        class PythonObjectProperties(NSObject):
+            object = objc_property(object)
+
+        class PythonObject:
+            pass
+
+        properties = PythonObjectProperties.alloc().init()
+
+        o = PythonObject()
+        wr = weakref.ref(o)
+
+        properties.object = o
+
+        # Test that Python object is properly stored.
+        self.assertIs(properties.object, o)
+
+        # Test that Python object is retained by the property.
+        del o
+        gc.collect()
+
+        self.assertIs(properties.object, wr())
+
+        # Test that Python object is released by the property.
+        properties.object = None
+        gc.collect()
+        self.assertIsNone(wr())
+
+        # Test that Python object is released by dealloc.
+
+        o = PythonObject()
+        wr = weakref.ref(o)
+
+        properties.object = o
+        self.assertIs(properties.object, o)
+
+        del o
+        del properties
+        gc.collect()
+
+        self.assertIsNone(wr())
+
+    def test_class_python_properties_weak(self):
+
+        class WeakPythonObjectProperties(NSObject):
+            object = objc_property(object, weak=True)
+
+        class PythonObject:
+            pass
+
+        properties = WeakPythonObjectProperties.alloc().init()
+
+        o = PythonObject()
+        wr = weakref.ref(o)
+
+        properties.object = o
+
+        # Test that Python object is properly stored.
+        self.assertIs(properties.object, o)
+
+        # Test that Python object is not retained by the property.
+        del o
+        gc.collect()
+
+        self.assertIsNone(properties.object)
+        self.assertIsNone(wr())
+
     def test_class_nonobject_properties(self):
         """An Objective-C class can have properties of non-object types."""
 
-        class Properties(NSObject):
+        class NonObjectProperties(NSObject):
             object = objc_property(ObjCInstance)
             int = objc_property(c_int)
             rect = objc_property(NSRect)
 
-        properties = Properties.alloc().init()
+        properties = NonObjectProperties.alloc().init()
 
         properties.object = at('foo')
         properties.int = 12345
@@ -997,6 +1068,51 @@ class RubiconTest(unittest.TestCase):
         self.assertEqual(r.size.width, 56)
         self.assertEqual(r.size.height, 78)
 
+    def test_class_nonobject_properties_weak(self):
+
+        with self.assertRaises(TypeError):
+            class WeakNonObjectProperties(NSObject):
+                int = objc_property(c_int, weak=True)
+
+    def test_class_properties_lifecycle_strong(self):
+
+        class StrongObjectProperties(NSObject):
+            object = objc_property(ObjCInstance)
+
+        with autoreleasepool():
+
+            properties = StrongObjectProperties.alloc().init()
+
+            obj = NSObject.alloc().init()
+            obj_pointer = obj.ptr.value  # store the object pointer for future use
+
+            properties.object = obj
+
+            del obj
+            gc.collect()
+
+        # assert that the object was retained by the property
+        self.assertEqual(properties.object.ptr.value, obj_pointer)
+
+    def test_class_properties_lifecycle_weak(self):
+
+        class WeakObjectProperties(NSObject):
+            object = objc_property(ObjCInstance, weak=True)
+
+        with autoreleasepool():
+
+            properties = WeakObjectProperties.alloc().init()
+
+            obj = NSObject.alloc().init()
+            properties.object = obj
+
+            self.assertIs(properties.object, obj)
+
+            del obj
+            gc.collect()
+
+        self.assertIsNone(properties.object)
+
     def test_class_with_wrapped_methods(self):
         """An ObjCClass can have wrapped methods."""
 
@@ -1089,7 +1205,7 @@ class RubiconTest(unittest.TestCase):
         class StructReturnHandler(Thing):
             @objc_method
             def initWithValue_(self, value):
-                self.value = value
+                self.value = py_from_ns(value)
                 return self
 
             @objc_method
@@ -1146,3 +1262,208 @@ class RubiconTest(unittest.TestCase):
         outSize = obj.testThing(15)
         self.assertEqual(outSize.width, 5)
         self.assertEqual(outSize.height, 45)
+
+    def test_objcinstance_python_attribute(self):
+        """Python attributes can be added to an ObjCInstance."""
+
+        Thing = ObjCClass("Thing")
+        thing = Thing.alloc().init()
+
+        # Use objects that don't have an obvious Objective-C equivalent,
+        # to ensure that the actual Python objects are being stored,
+        # and not converted Objective-C versions.
+        thing.python_object_1 = range(2, 8)
+        thing.python_object_2 = type
+        self.assertEqual(thing.python_object_1, range(2, 8))
+        self.assertEqual(thing.python_object_2, type)
+
+        # Test deleting Python attribute.
+
+        del thing.python_object_1
+
+        with self.assertRaises(AttributeError):
+            thing.python_object_1
+
+    def test_objcinstance_python_attribute_keep_alive(self):
+        """Python attributes on an ObjCInstance are kept even if the object temporarily has no Python references."""
+
+        Example = ObjCClass("Example")
+        example = Example.alloc().init()
+        Thing = ObjCClass("Thing")
+        thing = Thing.alloc().init()
+
+        # Use objects that don't have an obvious Objective-C equivalent,
+        # to ensure that the actual Python objects are being stored,
+        # and not converted Objective-C versions.
+        python_object_1 = range(2, 8)
+        python_object_2 = type
+
+        # Remember the objects' IDs to allow checking that the objects retrieved later are identical
+        # without keeping an actual reference to the objects.
+        python_object_1_id = id(python_object_1)
+        python_object_2_id = id(python_object_2)
+
+        # Add our Python attributes to the Objective-C object.
+        thing.python_object_1 = python_object_1
+        thing.python_object_2 = python_object_2
+
+        # Store the object in an Objective-C property.
+        # This creates a reference in Objective-C, but not in Python.
+        example.setThing(thing)
+
+        # Delete all of our Python references to the ObjCInstance and the objects stored on it.
+        del python_object_1
+        del python_object_2
+        del thing
+
+        # Try to force Python to destroy any no longer referenced objects.
+        gc.collect()
+
+        # Get our object back from Objective-C.
+        thing = example.thing
+
+        # Check that our Python attributes are still there.
+        self.assertEqual(thing.python_object_1, range(2, 8))
+        self.assertEqual(thing.python_object_2, type)
+
+        # Check that these are exactly the same objects that we stored before.
+        self.assertEqual(id(thing.python_object_1), python_object_1_id)
+        self.assertEqual(id(thing.python_object_2), python_object_2_id)
+
+    def test_objcinstance_python_attribute_freed(self):
+        """Python attributes on an ObjCInstance are freed after the instance is released."""
+
+        with autoreleasepool():
+
+            obj = NSObject.alloc().init()
+
+            # Use a custom object as attribute value so that we can keep a weak reference.
+
+            class TestO:
+                pass
+
+            python_object = TestO()
+
+            wr_python_object = weakref.ref(python_object)
+
+            # Add our Python attribute to the Objective-C object.
+            obj.python_object = python_object
+
+            # Delete all of Python references to the Python object.
+            del python_object
+
+            # Try to force Python to destroy any no longer referenced objects.
+            gc.collect()
+
+            # Check that our Python attributes are still there.
+            self.assertIs(obj.python_object, wr_python_object())
+
+            # Make sure that all Python objects are freed.
+            del obj
+            gc.collect()
+
+        self.assertIsNone(wr_python_object())
+
+    def test_objcinstance_release_owned(self):
+
+        # Create an object which we own.
+        obj = NSObject.alloc().init()
+
+        # Check that it is marked for release.
+        self.assertTrue(obj._needs_release)
+
+        # Explicitly release the object.
+        obj.release()
+
+        # Check that we no longer need to release it.
+        self.assertFalse(obj._needs_release)
+
+        # Delete it and make sure that we don't segfault on garbage collection.
+        del obj
+        gc.collect()
+
+    def test_objcinstance_autorelease_owned(self):
+
+        # Create an object which we own.
+        obj = NSObject.alloc().init()
+
+        # Check that it is marked for release.
+        self.assertTrue(obj._needs_release)
+
+        # Explicitly release the object.
+        res = obj.autorelease()
+
+        # Check that autorelease call returned the object itself.
+        self.assertIs(obj, res)
+
+        # Check that we no longer need to release it.
+        self.assertFalse(obj._needs_release)
+
+        # Delete it and make sure that we don't segfault on garbage collection.
+        del obj
+        gc.collect()
+
+    def test_objcinstance_retain_release(self):
+        NSString = ObjCClass('NSString')
+
+        # Create an object which we don't own.
+        string = NSString.stringWithString('test')
+
+        # Check that it is not marked for release.
+        self.assertFalse(string._needs_release)
+
+        # Explicitly retain the object.
+        res = string.retain()
+
+        # Check that autorelease call returned the object itself.
+        self.assertIs(string, res)
+
+        # Manually release the object.
+        string.release()
+
+        # Delete it and make sure that we don't segfault on garbage collection.
+        del string
+        gc.collect()
+
+    def test_objcinstance_dealloc(self):
+
+        class DeallocTester(NSObject):
+            attr0 = objc_property()
+            attr1 = objc_property(weak=True)
+
+            @objc_method
+            def dealloc(self):
+                self._did_dealloc = True
+
+        obj = DeallocTester.alloc().init()
+        obj.__dict__["_did_dealloc"] = False
+
+        attr0 = NSObject.alloc().init()
+        attr1 = NSObject.alloc().init()
+
+        obj.attr0 = attr0
+        obj.attr1 = attr1
+
+        self.assertEqual(attr0.retainCount(), 2)
+        self.assertEqual(attr1.retainCount(), 1)
+
+        # ObjC object will be deallocated, can only access Python attributes afterwards.
+        obj.release()
+
+        self.assertTrue(obj._did_dealloc, "custom dealloc did not run")
+        self.assertEqual(attr0.retainCount(), 1, "strong property value was not released")
+        self.assertEqual(attr1.retainCount(), 1, "weak property value was released")
+
+    def test_partial_with_override(self):
+        "If one method in a partial is overridden, that doesn't impact lookup of other partial targets"
+        SpecificExample = ObjCClass("SpecificExample")
+
+        obj = SpecificExample.alloc().init()
+
+        # The subclass implementation is invoked, not the base
+        obj.method(2, withArg=3)
+        self.assertEqual(obj.baseIntField, 5)
+
+        # The base class implementation can still be found an invoked.
+        obj.method(2)
+        self.assertEqual(obj.baseIntField, 2)
